from collections import OrderedDict
from functools import lru_cache
from typing import Tuple, List, Optional
from pathlib import Path
from random import randint, shuffle
import numpy as np
import skmultiflow as skm
from sklearn.kernel_ridge import KernelRidge
from scipy import stats
from dovado_rtl.enums import RegressionModel
from sklearn.model_selection import ShuffleSplit, RandomizedSearchCV
from dovado_rtl.config import Configuration
from dovado_rtl.abstract_classes import (
    AbstractEstimator,
    AbstractDesignPointEvaluator,
)
from dovado_rtl.simple_types import DesignValue, Example, Metric


class Estimator(AbstractEstimator):
    def __init__(
        self,
        regression_model: RegressionModel,
        design_point_evaluator: AbstractDesignPointEvaluator,
        free_parameters_range: "OrderedDict[str, Tuple[int, int]]",
        dataset_size: int,
        config: Configuration,
        testing_estimation: bool,
        recording_design_values: bool,
        reading_design_values: bool,
    ):
        self.__regression_model = regression_model
        self.__reg = None
        self.__last_design_point = None
        self.__last_len = None
        self.__examples: List[Example] = []
        self.__design_point_evaluator: AbstractDesignPointEvaluator = design_point_evaluator
        self.__dataset_size: int = dataset_size
        self.__free_parameters_range = free_parameters_range
        self.__config: Configuration = config
        self.__testing_estimation: bool = testing_estimation
        self.__recording_design_values = recording_design_values
        self.__reading_design_values = reading_design_values
        self.__design_point_evaluator.set_estimator(self)
        self.__metrics: Optional[List[Metric]] = None
        if testing_estimation:
            self.__estimation_fname = str(
                self.__config.get_config("WORK_DIR")
            ) + str(self.__config.get_config("EST_TEST_CSV"))
            Path(self.__estimation_fname).open("w").close()
            Path(self.__estimation_fname).open("a").write(
                "Real, Estimated, Metric\n"
            )
        if self.__recording_design_values:
            self.__sample_fname = str(
                self.__config.get_config("WORK_DIR")
            ) + str(self.__config.get_config("SAMPLE_RECORD_CSV"))
            Path(self.__sample_fname).open("w").close()
            top_line = ""
            for k in free_parameters_range.keys():
                top_line += k + ","
            Path(self.__sample_fname).open("a").write(top_line)

        self.__generate_dataset()

    @lru_cache()
    def estimate(self, design_point: Tuple[int, ...], metric: Metric) -> float:
        if self.__regression_model is RegressionModel.KERNEL_RIDGE:
            estimate = self.__kernel_ridge(list(design_point), metric)
        elif self.__regression_model in {
            RegressionModel.STREAM_LEARNING_ISOUP,
            RegressionModel.STREAM_LEARNING_STACKED,
        }:
            estimate = self.__stream_learner(list(design_point), metric)
        else:
            raise Exception(
                "Learinig method "
                + str(self.__regression_model)
                + " is not implemented"
            )

        if self.__testing_estimation and len(self.__examples) > max(
            len(list(design_point)), 5
        ):
            evaluated = self.__design_point_evaluator.evaluate(
                tuple(design_point)
            )
            line = (
                str(evaluated.value[metric])
                + ","
                + str(estimate)
                + ","
                + (
                    str(metric.utilisation[0])
                    + "-"
                    + str(metric.utilisation[1])
                    if not metric.is_frequency
                    else "Frequency"
                )
            )

            Path(self.__estimation_fname).open("a").writelines([line + "\n"])

        return estimate

    def __prepare_data(
        self, design_point: List[int], metric: Optional[Metric] = None,
    ):
        X = self.__get_independent_variables()
        X = X.astype("int64").reshape(-1, len(design_point))
        if metric:
            y = self.__get_dependent_variables(metric=metric)
            y = y.astype("float64").reshape(-1, 1)
        else:
            y = self.__get_dependent_variables(metric=metric)
            if self.__metrics:
                y = y.astype("float64").reshape(-1, len(self.__metrics))
            else:
                raise Exception(
                    "Metrics not set while building data for estimator"
                )
        return X, y

    def __stream_learner(
        self, design_point: List[int], metric: Metric
    ) -> float:
        X, y = None, None
        if self.__last_design_point != design_point:
            X, y = self.__prepare_data(design_point)
        if not self.__last_design_point:
            if (
                self.__regression_model
                is RegressionModel.STREAM_LEARNING_ISOUP
            ):
                self.__reg = skm.trees.iSOUPTreeRegressor()
            elif (
                self.__regression_model
                is RegressionModel.STREAM_LEARNING_STACKED
            ):
                self.__reg = (
                    skm.trees.StackedSingleTargetHoeffdingTreeRegressor()
                )
            else:
                Exception("Streaming regression model not implemented")
            self.__reg.fit(X, y)
            self.__last_design_point = design_point
            self.__last_len = len(X)
        elif self.__last_design_point != design_point:
            if self.__last_len != len(X):
                start_index = -(self.__last_len - len(X))
                Xl = X[start_index:]
                yl = y[start_index:]
                Xl = Xl.astype("int64").reshape(-1, len(design_point))
                if self.__metrics:
                    yl = yl.astype("float64").reshape(-1, len(self.__metrics))
                else:
                    raise Exception(
                        "Metrics not initialized while partially fitting streaming learner"
                    )
                self.__reg.partial_fit(Xl, yl)
                self.__last_len = len(X)
            self.__last_design_point = design_point
        design_point = np.reshape(design_point, (1, -1))
        return self.__reg.predict(design_point)[0][
            self.__metrics.index(metric)
        ]

    def __kernel_ridge(self, design_point: List[int], metric: Metric) -> float:
        X, y = None, None
        if self.__last_design_point != design_point:
            X, y = self.__prepare_data(design_point)

        if (
            self.__last_design_point != design_point
            and self.__last_len != len(X)
        ):
            self.__reg = KernelRidge()
            parameters = {
                "alpha": stats.loguniform(a=1e-4, b=1e2),
                "gamma": stats.expon(scale=0.1),
                "kernel": ["rbf"],
            }
            cv = ShuffleSplit(n_splits=10, test_size=0.2, random_state=0)
            self.__reg = RandomizedSearchCV(
                self.__reg,
                parameters,
                cv=cv,
                n_iter=50,
                scoring="neg_root_mean_squared_error",
                n_jobs=-1,
            )
            self.__reg.fit(X, y)
            self.__last_design_point = design_point
            self.__last_len = len(X)
        design_point = np.reshape(design_point, (1, -1))
        return self.__reg.predict(design_point)[0][
            self.__metrics.index(metric)
        ]

    def add_example(self, example: Example) -> None:
        if not self.__metrics:
            self.__metrics = self.__design_point_evaluator.get_metrics()
        self.__examples.append(example)
        shuffle(self.__examples)
        self.__examples_updated = True
        if self.__recording_design_values:
            opened_file = Path(self.__sample_fname).open("a")
            if Path(self.__sample_fname).read_text()[-1] == ",":
                top_line = ""
                for m in example.design_value.value.keys():
                    top_line += (
                        m.utilisation[0] + "-" + m.utilisation[1]
                        if not m.is_frequency
                        else "Frequency"
                    ) + ","
                opened_file.write(top_line[:-1] + "\n")
            line = ""
            for i in example.design_point:
                line += str(i) + ","
            for j in example.design_value.value.values():
                line += str(j) + ","
            opened_file.write(line[:-1] + "\n")
            opened_file.close()

    @staticmethod
    def evaluate_random_design_point(
        free_parameters_range: "OrderedDict[str, Tuple[int, int]]",
        evaluator: AbstractDesignPointEvaluator,
    ) -> Tuple[List[int], DesignValue]:

        design_point = []
        for k in free_parameters_range.keys():
            design_point.append(
                randint(
                    free_parameters_range[k][0], free_parameters_range[k][1],
                )
            )
        return design_point, evaluator.evaluate(tuple(design_point))

    def __generate_dataset(self) -> None:
        design_value = None
        if self.__reading_design_values:
            lines = (
                Path(
                    str(self.__config.get_config("WORK_DIR"))
                    + str(self.__config.get_config("SAMPLE_RECORD_CSV"))
                )
                .read_text()
                .split("\n")
            )
            splitted_top_line = lines[0].split(",")
            metrics_to_set = True
            metrics = []
            for line in lines[1:]:
                try:
                    design_point = []
                    splitted_line = line.split(",")
                    i = 0
                    for _ in self.__free_parameters_range.keys():
                        design_point.append(float(splitted_line[i]))
                        i += 1

                    design_value = DesignValue({})
                    for v in splitted_line[i:]:
                        if splitted_top_line[i] == "Frequency":
                            metric = Metric(None, True)
                        else:
                            util = splitted_top_line[i].split("-")
                            metric = Metric((util[0], util[1]), False)
                        if metrics_to_set:
                            metrics.append(metric)
                        design_value.value[metric] = float(v)
                        i += 1
                    if metrics_to_set:
                        self.__design_point_evaluator.set_metrics(metrics)
                        metrics_to_set = False

                    self.add_example(
                        Example(
                            design_point=design_point,
                            design_value=design_value,
                        )
                    )
                except Exception:
                    print(
                        "Error while parsing line '" + str(line) + "' skipping"
                    )
            return

        for _ in range(0, self.__dataset_size):
            (
                design_point,
                design_value,
            ) = Estimator.evaluate_random_design_point(
                self.__free_parameters_range, self.__design_point_evaluator
            )
            if design_value:
                self.add_example(
                    Example(
                        design_point=design_point, design_value=design_value
                    )
                )
        if not design_value:
            raise Exception(
                "Could not find any valid design point while sampling from the given ranges"
                + " please make sure the ranges are correct. You may try again or modify dataset size if ranges are correct."
            )

    def __get_dependent_variables(
        self, metric: Optional[Metric] = None
    ) -> "np.ndarray":
        dependent_variable = []
        if not metric:
            for example in self.__examples:
                dependent_variable.append(
                    list(example.design_value.value.values())
                )
        else:
            for example in self.__examples:
                dependent_variable.append(example.design_value.value[metric])
        return np.array(dependent_variable)

    def __get_independent_variables(self) -> "np.ndarray":
        independent_variables = []
        for example in self.__examples:
            independent_variables.append(example.design_point)
        return np.array(independent_variables)

    def get_examples(self):
        return self.__examples



import pexpect
import threading
import re
from typing import Tuple, Dict

vivado = None
t_vivado_init = None


def start():
    global vivado
    global t_vivado_init
    vivado = pexpect.spawnu(
        "vivado -mode tcl -nolog -nojournal",
        encoding="utf-8",
        codec_errors="ignore",
    )

    t_vivado_init = threading.Thread(target=vivado.expect, args=("Vivado%",))
    t_vivado_init.start()
    return vivado


def execute_command(command: str) -> str:
    if not vivado.isalive():
        raise Exception(
            "Could not start Vivado, is it installed?"
            + " Is it in your PATH?\n"
            + " Try running 'vivado -mode tcl' yourself and"
            + " debug why it does not start"
        )
    global t_vivado_init
    if t_vivado_init is not None:
        print("Waiting for Vivado to start ...")
        t_vivado_init.join()
        vivado.sendline("set_param tcl.collectionResultDisplayLimit Inf")
        vivado.expect("Vivado%", timeout=None)
        t_vivado_init = None

    vivado.sendline(command)
    vivado.expect("Vivado%", timeout=None)
    return str(vivado.before)


def source(tcl_script: str) -> Tuple[str, bool]:
    vivado_out = execute_command("source " + tcl_script)
    return vivado_out, not re.findall("ERROR:", vivado_out)


def get_parts() -> Dict[str, str]:
    parts = [
        x.strip()
        for x in re.sub(
            "get_parts\r\n", "", execute_command("get_parts")
        ).split(" ")[1:]
    ]
    family = [
        x.strip()
        for x in re.sub(
            "[get_parts]\r\n",
            "",
            execute_command("get_property FAMILY [get_parts]"),
        ).split(" ")[3:]
    ]
    family[0] = family[0][len("[get_parts]\r\n") :]
    return dict(zip(parts, family))


def get_help(command: str) -> str:
    return execute_command("help " + command)



from typing import List, Dict, Tuple, Optional, Any
from dataclasses import dataclass


@dataclass(eq=True, frozen=True)  # this makes Metric hashable
class Metric:
    utilisation: Optional[Tuple[str, str]]
    is_frequency: bool


@dataclass
class IsIncremental:
    synthesis: bool
    implementation: bool


@dataclass
class DesignValue:
    value: Dict[Metric, float]


@dataclass
class Example:
    design_point: List[int]
    design_value: DesignValue



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dovado_rtl',
 'dovado_rtl.antlr',
 'dovado_rtl.antlr.generated',
 'dovado_rtl.antlr.generated.SysVerilogHDL',
 'dovado_rtl.antlr.generated.Verilog2001',
 'dovado_rtl.antlr.generated.vhdl',
 'tcl',
 'verilog',
 'vhdl',
 'xdc']

package_data = \
{'': ['*'], 'dovado_rtl.antlr': ['grammars/*', 'grammars/.antlr/*']}

install_requires = \
['BeautifulSoup4>=4.9.1,<5.0.0',
 'antlr4-python3-runtime>=4.8.0,<4.9.0',
 'click>=7.1.2,<8.0.0',
 'importlib-resources>=3.3.0,<4.0.0',
 'lxml>=4.5.2,<5.0.0',
 'numpy>=1.19.2,<2.0.0',
 'pathvalidate>=2.3.0,<3.0.0',
 'pexpect>=4.8.0,<5.0.0',
 'pymoo>=0.4.2,<0.5.0',
 'pyyaml>=5.3.1,<6.0.0',
 'scikit-multiflow>=0.5.3,<0.6.0',
 'scipy>=1.5.4,<2.0.0',
 'sklearn>=0.0,<0.1',
 'typer>=0.3.2,<0.4.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.8,<0.9']}

entry_points = \
{'console_scripts': ['dovado = dovado_rtl.main:main']}

setup_kwargs = {
    'name': 'dovado-rtl',
    'version': '0.3.9',
    'description': 'CLI tool for RTL Design Space Exploration on top of Vivado',
    'long_description': '\n\n# DoVado\n\nAdvanced Computer Architectures (ACA) Research project 2020.   \nA CLI tool for RTL Design Space Exploration on top of Vivado.   \n&ldquo;Do&rsquo; vado?&rdquo; is an italian slang expression for &ldquo;Dove vado?&rdquo; which means &ldquo;Where do I go?&rdquo;\n\n\n# State of the Project\n\nComplete prototype.\n\n\n# How to inspect the project\n\nThis project uses [Poetry](https://python-poetry.org/) for managing dependences and python versions in order to avoid conflicting versions on different machines.\nFollowing instructions are tested on a Linux machine but should work on OSX with minor (or none) modifications.\n\n\n## Install Poetry\n\n(if you already have poetry installed skip to the next section)   \nExecute the following command in a shell to install poetry\n\n    curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python\n\nYou will find poetry in $HOME/poetry/.bin where $HOME is usually something like *home/your-user-name* (on Linux).\nTo check that everything went well run:\n\n    poetry --version\n\nIf this does not work add the poetry installation directory to the PATH ([instructions](https://docs.oracle.com/cd/E19062-01/sun.mgmt.ctr36/819-5418/gaznb/index.html) for Linux)\n\n\n## Run an example\n\nClone this repository to your local machine:\n\n    git clone  --recurse-submodules -j8 https://github.com/DPaletti/dovado.git\n\nNow position at the project root (from now on all commands assume you are at project root):\n\n    cd dovado\n\nInstall all the required dependences:\n\n    poetry install\n\nRun an example:\n\n    poetry run dovado < examples/input_files/input_rtl_vadd.txt\n\nDoing so all the program prompts are automatically answered with each line in input (one line = one answer), open it to see the answers.\nAfter all the vivado output you should see examples&rsquo; WNS (worst negative slack) and LUT (lookup table) percentage utilization\n\n\n# Testing\n\nIn order to run tests:\n\n    poetry run pytest\n\nAll tests will be ran and their respective outcome shown.   \nTests are managed through pytest which is used as a testing library and as a test runner. Functions which call Vivado are mocked through monkeypatch (from pytest).   \n\nA recap of test coverage can be read by:\n\n    cd html_cov/\n    firefox index.html\n\nif you do not have firefox installed any other browser will do.   \n\n\n# Report\nA full report of the activity conducted in developing Dovado and studying the RTL design space exploration problem can be read ([here](./dovado.pdf)).\n',
    'author': 'Daniele Paletti',
    'author_email': 'danielepaletti98@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/DPaletti/dovado',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
